import datetime
import logging
import random
import threading
import os
import time

import boto3
import pydantic
import obsv_tools.metrics.instrumentator
import sqs_worker


logger = logging.getLogger('sqs_worker_example')

logging.basicConfig(level=logging.INFO)


class MessagePayload(
    pydantic.BaseModel,
):
    value: str


class MyWorker(
    sqs_worker.worker.Worker,
):

    def __init__(
        self,
        sqs_client,
        queue_name: str,
        metrics_instrumentator: obsv_tools.metrics.instrumentator.Instrumentator,
    ) -> None:
        super().__init__(
            logger=logger,
            name='MyWorker',
            sqs_client=sqs_client,
            queue_name=queue_name,
            metrics_instrumentator=metrics_instrumentator,
            payload_class=MessagePayload,
            max_messages=2,
            idle_limit=datetime.timedelta(minutes=10),
        )

    def work(
        self,
        messages: list[sqs_worker.models.Message[MessagePayload]],
    ) -> None:
        thread_id = threading.get_ident()
        for message in messages:
            wait_time = random.uniform(0.1, 9.5)
            time.sleep(wait_time)

            logger.info(
                msg=f'Thread {thread_id}: Finished processing message with: {message.payload.value}',
            )


def wait_for_sqs_queue(
    sqs_client,
    queue_name: str,
) -> None:
    max_retries = 30
    for i in range(max_retries):
        try:
            sqs_client.list_queues()

            sqs_client.get_queue_url(QueueName=queue_name)
            logger.info(f'Queue {queue_name} is available')

            return
        except sqs_client.exceptions.QueueDoesNotExist:
            logger.info(
                msg=f'Queue {queue_name} does not exist yet, waiting...',
                extra={
                    'queue_name': queue_name,
                    'attempt': i + 1,
                    'max_retries': max_retries,
                },
            )
            time.sleep(2)
        except Exception:
            logger.info(
                msg='Waiting for SQS service...',
                extra={
                    'queue_name': queue_name,
                    'attempt': i + 1,
                    'max_retries': max_retries,
                },
            )
            time.sleep(2)

    raise Exception(f'Queue {queue_name} is not available after waiting')


def main() -> None:
    # Thread safe clients only
    metrics_instrumentator = obsv_tools.metrics.instrumentator.Instrumentator(
        server_port=8000,
    )

    sqs_client = boto3.client(
        'sqs',
        endpoint_url=os.environ['SQS_ENDPOINT_URL'],
        region_name=os.environ.get('AWS_REGION', 'us-east-1'),
        aws_access_key_id=os.environ.get('AWS_ACCESS_KEY_ID', 'test'),
        aws_secret_access_key=os.environ.get('AWS_SECRET_ACCESS_KEY', 'test'),
    )
    wait_for_sqs_queue(
        sqs_client=sqs_client,
        queue_name=os.environ['SQS_QUEUE_NAME'],
    )

    def worker_factory():
        return MyWorker(
            sqs_client=boto3.client(
                'sqs',
                endpoint_url=os.environ['SQS_ENDPOINT_URL'],
                region_name=os.environ.get('AWS_REGION', 'us-east-1'),
                aws_access_key_id=os.environ.get('AWS_ACCESS_KEY_ID', 'test'),
                aws_secret_access_key=os.environ.get('AWS_SECRET_ACCESS_KEY', 'test'),
            ),
            queue_name=os.environ['SQS_QUEUE_NAME'],
            metrics_instrumentator=metrics_instrumentator,
        )

    sqs_worker.worker.multiple(
        worker_factory=worker_factory,
        n=1,
    )


if __name__ == '__main__':
    main()


import datetime
import logging
import threading

import boto3
import mypy_boto3_sqs.client
import pydantic
import obsv_tools.metrics.instrumentator
import sqs_worker

import sqs_utils


logger = logging.getLogger('sqs_worker_example')

logging.basicConfig(level=logging.INFO)


class MessagePayload(
    pydantic.BaseModel,
):
    value: str


class MyWorker(
    sqs_worker.worker.Worker,
):

    def __init__(
        self,
        sqs_client: mypy_boto3_sqs.client.SQSClient,
        queue_name: str,
        metrics_instrumentator: obsv_tools.metrics.instrumentator.Instrumentator,
    ) -> None:
        super().__init__(
            logger=logger,
            name='MyWorker',
            sqs_client=sqs_client,
            queue_name=queue_name,
            metrics_instrumentator=metrics_instrumentator,
            payload_class=MessagePayload,
            max_messages=2,
            idle_limit=datetime.timedelta(seconds=2),
        )

    def work(
        self,
        messages: list[sqs_worker.models.Message[MessagePayload]],
    ) -> None:
        thread_id = threading.get_ident()
        for message in messages:
            logger.info(
                msg=f'Thread {thread_id}: Processing message with: {message.payload.value}',
            )


def main() -> None:
    # Thread safe clients only
    metrics_instrumentator = obsv_tools.metrics.instrumentator.Instrumentator(
        server_port=9090,
    )

    sqs_client = boto3.client(
        'sqs',
        endpoint_url='http://localhost:4566',
        region_name='us-east-1',
        aws_access_key_id='test',
        aws_secret_access_key='test',
    )
    queue_name = 'my-queue'
    number_of_workers = 10

    if not sqs_utils.sqs_connected(sqs_client=sqs_client):
        raise RuntimeError('Cannot connect to SQS. Please ensure LocalStack is running by executing "localstack start".')

    sqs_utils.ensure_queue_exists(
        sqs_client=sqs_client,
        queue_name=queue_name,
    )

    for _ in range(number_of_workers):
        sqs_utils.send_hello_message(
            sqs_client=sqs_client,
            queue_name=queue_name,
            message=sqs_worker.models.Message(
                version='1.0',
                msg_type='greeting',
                payload=MessagePayload(value='Hello, World!'),
            ),
    )

    def worker_factory():
        return MyWorker(
            sqs_client=boto3.client(
                'sqs',
                endpoint_url='http://localhost:4566',
                region_name='us-east-1',
                aws_access_key_id='test',
                aws_secret_access_key='test',
            ),
            queue_name=queue_name,
            metrics_instrumentator=metrics_instrumentator,
        )

    sqs_worker.worker.multiple(
        worker_factory=worker_factory,
        n=number_of_workers,
    )


if __name__ == '__main__':
    main()


#!/usr/bin/env python3
"""
Script to send test messages to the SQS queue to generate metrics.
Run this after starting the docker-compose stack.
"""

import boto3
import json
import time
import random
import sys

def main():
    # Configure SQS client for LocalStack
    sqs_client = boto3.client(
        'sqs',
        endpoint_url='http://localhost:4566',
        region_name='us-east-1',
        aws_access_key_id='test',
        aws_secret_access_key='test',
    )
    
    queue_name = 'worker-queue'
    
    try:
        # Get queue URL
        response = sqs_client.get_queue_url(QueueName=queue_name)
        queue_url = response['QueueUrl']
        print(f"Found queue: {queue_url}")
    except Exception as e:
        print(f"Error getting queue URL: {e}")
        print("Make sure the docker-compose stack is running and the queue is created.")
        sys.exit(1)
    
    # Send test messages
    message_count = 50
    print(f"Sending {message_count} test messages...")
    
    for i in range(message_count):
        message_body = json.dumps({
            "version": "1.0",
            "msg_type": "test",
            "payload": {
                "value": f"Test message {i + 1}"
            }
        })
        
        try:
            sqs_client.send_message(
                QueueUrl=queue_url,
                MessageBody=message_body
            )
            print(f"Sent message {i + 1}/{message_count}")
            
            # Random delay between messages
            time.sleep(random.uniform(0.1, 0.5))
            
        except Exception as e:
            print(f"Error sending message {i + 1}: {e}")
    
    print("Finished sending messages!")
    print("\nYou can now check the metrics at:")
    print("- Prometheus: http://localhost:9091")
    print("- Grafana: http://localhost:3000 (admin/admin)")

if __name__ == "__main__":
    main()



import concurrent.futures
import datetime
import itertools
import json
import logging
import signal
import time
import typing

import obsv_tools.metrics.instrumentator

from . import exceptions
from . import models

import mypy_boto3_sqs.client


MAX_BATCH_SIZE = 10


class Worker:
    EXCEPTIONS_METRIC_NAME = 'sqs.worker.exceptions'
    WORK_MESSAGES_METRIC_NAME = 'sqs.worker.work.messages'
    WORK_LATENCY_METRIC_NAME = 'sqs.worker.work.latency'

    def __init__(
        self,
        logger: logging.Logger,
        name: str,
        sqs_client: mypy_boto3_sqs.client.SQSClient,
        queue_name: str,
        payload_class: type[models.T],
        max_messages: int = MAX_BATCH_SIZE,
        metrics_instrumentator: obsv_tools.metrics.instrumentator.Instrumentator | None = None,
        wait_for_max_messages: datetime.timedelta = datetime.timedelta(seconds=1),
        visibility_timeout: int = 30,
        auto_ack: bool = True,
        max_retry_attempts: int = 10,
        pull_msgs_interval: datetime.timedelta = datetime.timedelta(seconds=0.5),
        idle_limit: datetime.timedelta | None = None,
    ) -> None:
        self.name = name
        self.sqs_client = sqs_client
        self.queue_name = queue_name
        self.logger = logger
        self.max_messages = max_messages
        self.wait_for_messages = wait_for_max_messages
        self.visibility_timeout = visibility_timeout
        self.active = True
        self.metrics_instrumentator = metrics_instrumentator
        self.payload_class = payload_class
        self.auto_ack = auto_ack
        self.max_retry_attempts = max_retry_attempts
        self.pull_msgs_interval = pull_msgs_interval
        self.idle_limit = idle_limit

        self.metrics_instrumentator.add_counter(
            name=self.EXCEPTIONS_METRIC_NAME,
            description='Exception Counter',
        ).add_counter(
            name=self.WORK_MESSAGES_METRIC_NAME,
            description='Number of messages processed',
        ).add_histogram(
            name=self.WORK_LATENCY_METRIC_NAME,
            description='Time spent working on message',
        )

    def start(
        self,
    ) -> None:
        self.logger.info(
            msg='Starting worker',
            extra={
                'worker_name': self.name,
            },
        )

        idle_time = 0

        while self.active:
            try:
                raw_messages: list[mypy_boto3_sqs.type_defs.MessageTypeDef] = self.pull_messages()

                if not raw_messages:
                    time.sleep(self.pull_msgs_interval.total_seconds())

                    idle_time += self.pull_msgs_interval.total_seconds()
                    if self.idle_limit and idle_time >= self.idle_limit.total_seconds():
                        self.logger.info(
                            msg='Idle limit reached, stopping worker',
                            extra={
                                'worker_name': self.name,
                            },
                        )

                        self.stop()

                    continue

                messages = [
                    self.parse_message(
                        message=raw_message,
                    )
                    for raw_message in raw_messages
                ]
            except Exception:
                self.logger.exception(
                    msg='Failed to pull messages',
                )

                continue

            try:
                start_time = time.time()

                self.work(
                    messages=messages,
                )

            except exceptions.RetryLaterException:
                self.retry_later(
                    messages=messages,
                )
            except Exception as exception:
                self.logger.exception(
                    msg='Failed to process messages',
                    extra={
                        'worker_name': self.name,
                    },
                )

                self.metrics_instrumentator.increment_counter(
                    name=self.EXCEPTIONS_METRIC_NAME,
                    attributes={
                        'worker_name': self.name,
                        'error_type': type(exception).__name__,
                    },
                )

                self.on_error(
                    exception=exception,
                    messages=messages,
                )
            else:
                self.logger.info(
                    msg='Worker successfully processed messages',
                    extra={
                        'worker_name': self.name,
                    },
                )

                self.metrics_instrumentator.increment_counter(
                    name=self.WORK_MESSAGES_METRIC_NAME,
                    attributes={
                        'worker_name': self.name,
                    },
                    amount=len(messages),
                )

                self.metrics_instrumentator.record_histogram(
                    name=self.WORK_LATENCY_METRIC_NAME,
                    attributes={
                        'worker_name': self.name,
                    },
                    amount=(time.time() - start_time),
                )

                if self.auto_ack:
                    self.ack_all(
                        messages=messages,
                    )

    def stop(
        self,
    ) -> None:
        self.logger.info(
            msg='Stopping worker',
            extra={
                'worker_name': self.name,
            },
        )

        self.active = False

    def work(
        self,
        messages: list[models.Message],
    ) -> None:
        raise NotImplementedError()

    def pull_messages(
        self,
    ) -> list[mypy_boto3_sqs.type_defs.MessageTypeDef]:
        queue_url = self.get_queue_url(
            queue_name=self.queue_name,
        )

        start_time = time.time()
        wait_time_sec = 0.0

        messages: list[mypy_boto3_sqs.type_defs.MessageTypeDef] = []

        wait_for_messages_sec = int(self.wait_for_messages.total_seconds())

        while len(messages) < self.max_messages and wait_time_sec <= wait_for_messages_sec:
            response = self.sqs_client.receive_message(
                QueueUrl=queue_url,
                MaxNumberOfMessages=min(10, self.max_messages - len(messages)),
                WaitTimeSeconds=max(1, min(20, wait_for_messages_sec - int(wait_time_sec))),
                VisibilityTimeout=self.visibility_timeout,
            )

            if 'Messages' in response:
                messages.extend(response['Messages'])

            wait_time_sec = time.time() - start_time

        return messages

    def parse_message(
        self,
        message: mypy_boto3_sqs.type_defs.MessageTypeDef,
    ) -> models.Message:
        message_body = json.loads(message['Body'])

        message_body.pop('queue_name', None)
        message_body.pop('receipt_handle', None)

        payload_data = message_body.pop('payload')
        payload = self.payload_class(**payload_data)

        return models.Message(
            queue_name=self.queue_name,
            receipt_handle=message['ReceiptHandle'],
            payload=payload,
            **message_body,
        )

    def retry_later(
        self,
        messages: list[models.Message],
    ):
        retry_messages = [
            message
            for message in messages
            if message.retry_attempt <= self.max_retry_attempts
        ]

        self.logger.info(
            msg='Enqueuing retry later messages',
            extra={
                'messages_count': len(retry_messages),
            },
        )

        for message in retry_messages:
            message.retry_attempt += 1

        self.publish_messages(
            queue_name=self.queue_name,
            messages=retry_messages,
            delay_seconds=datetime.timedelta(minutes=10),
        )

        self.ack_all(
            messages=retry_messages,
        )

    def publish_message(
        self,
        queue_name: str,
        message: models.Message,
    ) -> None:
        queue_url = self.get_queue_url(
            queue_name=queue_name,
        )

        self.sqs_client.send_message(
            QueueUrl=queue_url,
            MessageBody=message.model_dump_json(),
        )

    def publish_messages(
        self,
        queue_name: str,
        messages: list[models.Message],
        delay_seconds: datetime.timedelta | None = None,
    ) -> None:
        if not messages:
            return

        queue_url = self.get_queue_url(
            queue_name=queue_name,
        )

        for i in range(0, len(messages), MAX_BATCH_SIZE):
            batch = messages[i : i + MAX_BATCH_SIZE]

            json_messages: list[mypy_boto3_sqs.type_defs.SendMessageBatchRequestEntryTypeDef] = []

            for idx, message in enumerate(batch):
                entry = {
                    'Id': f'message-{idx}',
                    'MessageBody': message.model_dump_json(),
                }
                if delay_seconds is not None:
                    entry['DelaySeconds'] = int(delay_seconds.total_seconds())

                json_messages.append(entry)

            self.sqs_client.send_message_batch(
                QueueUrl=queue_url,
                Entries=json_messages,
            )

    def ack(
        self,
        message: models.Message,
    ) -> None:
        self.ack_all(
            messages=[
                message,
            ],
        )

    def ack_all(
        self,
        messages: list[models.Message],
    ) -> None:
        for messages_chunk in itertools.batched(messages, MAX_BATCH_SIZE):
            self.sqs_client.delete_message_batch(
                QueueUrl=self.get_queue_url(
                    queue_name=self.queue_name,
                ),
                Entries=[
                    {
                        'Id': f'message-{idx}',
                        'ReceiptHandle': message.receipt_handle,
                    }
                    for idx, message in enumerate(messages_chunk)
                    if message.receipt_handle is not None
                ],
            )

    def extend_message_visibility(
        self,
        message: models.Message,
        visibility_timeout: int,
    ):
        if message.receipt_handle is None:
            raise ValueError('Unable to extend visibility without receipt_handle')

        self.sqs_client.change_message_visibility(
            QueueUrl=self.get_queue_url(
                queue_name=self.queue_name,
            ),
            ReceiptHandle=message.receipt_handle,
            VisibilityTimeout=visibility_timeout,
        )

    def get_queue_url(
        self,
        queue_name: str,
    ) -> str:
        queue_data = self.sqs_client.get_queue_url(
            QueueName=queue_name,
        )

        return queue_data['QueueUrl']

    def on_error(
        self,
        exception: Exception,
        messages: list[models.Message],
    ) -> None:
        raise exception()


def single(
    worker_factory,
):
    multiple(
        worker_factory=worker_factory,
        n=1,
    )


def multiple(
    worker_factory: typing.Callable[[], Worker],
    n,
):
    pool = concurrent.futures.ThreadPoolExecutor(n)
    workers = [worker_factory() for _ in range(n)]

    def sigterm_handler(
        _signo,
        _stack_frame,
    ):
        for worker in workers:
            worker.stop()

    signal.signal(signal.SIGTERM, sigterm_handler)

    for worker in workers:
        pool.submit(worker.start)


import typing

import pydantic


T = typing.TypeVar('T', bound=pydantic.BaseModel)


class Message(
    pydantic.BaseModel,
    typing.Generic[T],
):
    version: str
    msg_type: str
    queue_name: str | None = None
    receipt_handle: str | None = None
    payload: T = pydantic.Field(
        repr=False,
    )
    retry_attempt: int = 0


import datetime
import json
import logging
import signal
import time
from unittest.mock import Mock, patch, ANY

import pytest
import pydantic

import sqs_worker.worker as worker
import sqs_worker.models as models
import sqs_worker.exceptions as exceptions
import obsv_tools.metrics.instrumentator as instrumentator


class PayloadForTesting(pydantic.BaseModel):
    message: str
    count: int


@pytest.fixture
def mock_sqs_client():
    return Mock()


@pytest.fixture
def mock_logger():
    return Mock(spec=logging.Logger)


@pytest.fixture
def mock_instrumentator():
    mock_inst = Mock(spec=instrumentator.Instrumentator)
    mock_inst.add_counter.return_value = mock_inst
    mock_inst.add_histogram.return_value = mock_inst
    return mock_inst


@pytest.fixture
def sample_payload():
    return PayloadForTesting(message="test message", count=42)


@pytest.fixture
def sample_message(
    sample_payload,
):
    return models.Message(
        version="1.0",
        msg_type="test",
        queue_name="test_queue",
        receipt_handle="receipt_123",
        payload=sample_payload,
        retry_attempt=0,
    )


@pytest.fixture
def sample_raw_message(
    sample_payload,
):
    return {
        'Body': json.dumps({
            'version': '1.0',
            'msg_type': 'test',
            'queue_name': 'test_queue',
            'receipt_handle': 'receipt_123',
            'payload': sample_payload.model_dump(),
            'retry_attempt': 0,
        }),
        'ReceiptHandle': 'receipt_123',
    }


@pytest.fixture
def test_worker(
    mock_logger,
    mock_sqs_client,
    mock_instrumentator,
):
    return worker.Worker(
        logger=mock_logger,
        name="test_worker",
        sqs_client=mock_sqs_client,
        queue_name="test_queue",
        payload_class=PayloadForTesting,
        max_messages=5,
        metrics_instrumentator=mock_instrumentator,
        wait_for_max_messages=datetime.timedelta(seconds=2),
        visibility_timeout=30,
        auto_ack=True,
        max_retry_attempts=3,
    )


class TestWorkerInit:
    def test___init___When_called_Then_sets_attributes(
        self,
        mock_logger,
        mock_sqs_client,
        mock_instrumentator,
    ):
        test_worker = worker.Worker(
            logger=mock_logger,
            name="test_worker",
            sqs_client=mock_sqs_client,
            queue_name="test_queue",
            payload_class=PayloadForTesting,
            max_messages=5,
            metrics_instrumentator=mock_instrumentator,
            wait_for_max_messages=datetime.timedelta(seconds=2),
            visibility_timeout=30,
            auto_ack=True,
            max_retry_attempts=3,
        )
        
        assert test_worker.name == "test_worker"
        assert test_worker.sqs_client == mock_sqs_client
        assert test_worker.queue_name == "test_queue"
        assert test_worker.logger == mock_logger
        assert test_worker.max_messages == 5
        assert test_worker.wait_for_messages == datetime.timedelta(seconds=2)
        assert test_worker.visibility_timeout == 30
        assert test_worker.active is True
        assert test_worker.metrics_instrumentator == mock_instrumentator
        assert test_worker.payload_class == PayloadForTesting
        assert test_worker.auto_ack is True
        assert test_worker.max_retry_attempts == 3

    def test___init___When_called_Then_configures_metrics(
        self,
        mock_instrumentator,
    ):
        worker.Worker(
            logger=Mock(),
            name="test_worker",
            sqs_client=Mock(),
            queue_name="test_queue",
            payload_class=PayloadForTesting,
            metrics_instrumentator=mock_instrumentator,
        )
        
        mock_instrumentator.add_counter.assert_called_once_with(
            counter_name='sqs.manager.exceptions',
            description='Exception Counter',
        )
        mock_instrumentator.add_histogram.assert_called_once_with(
            histogram_name='sqs.manager.work.latency',
            description='Time spent working on message',
        )


class TestWorkerStart:
    def test_start_When_no_messages_Then_sleeps_and_continues(
        self,
        test_worker,
    ):
        test_worker.pull_messages = Mock(return_value=[])
        
        # Mock the start method to avoid infinite loop
        def mock_start():
            raw_messages = test_worker.pull_messages()
            if not raw_messages:
                with patch('time.sleep') as mock_sleep:
                    time.sleep(0.5)
                    mock_sleep.assert_called_with(0.5)
        
        mock_start()

    def test_start_When_messages_available_Then_processes_messages(
        self,
        test_worker,
        sample_raw_message,
        sample_message,
    ):
        test_worker.pull_messages = Mock(return_value=[sample_raw_message])
        test_worker.parse_message = Mock(return_value=sample_message)
        test_worker.work = Mock()
        test_worker.ack_all = Mock()
        
        # Simulate one iteration of the start loop
        raw_messages = test_worker.pull_messages()
        messages = [test_worker.parse_message(message=raw_message) for raw_message in raw_messages]
        test_worker.work(messages=messages)
        if test_worker.auto_ack:
            test_worker.ack_all(messages=messages)
        
        test_worker.work.assert_called_once_with(messages=[sample_message])
        test_worker.ack_all.assert_called_once_with(messages=[sample_message])

    def test_start_When_work_raises_retry_later_exception_Then_retries_later(
        self,
        test_worker,
        sample_raw_message,
        sample_message,
    ):
        test_worker.pull_messages = Mock(return_value=[sample_raw_message])
        test_worker.parse_message = Mock(return_value=sample_message)
        test_worker.work = Mock(side_effect=exceptions.RetryLaterException())
        test_worker.retry_later = Mock()
        
        # Simulate one iteration of the start loop with exception handling
        raw_messages = test_worker.pull_messages()
        messages = [test_worker.parse_message(message=raw_message) for raw_message in raw_messages]
        try:
            test_worker.work(messages=messages)
        except exceptions.RetryLaterException:
            test_worker.retry_later(messages=messages)
        
        test_worker.retry_later.assert_called_once_with(messages=[sample_message])

    def test_start_When_work_raises_exception_Then_handles_error(
        self,
        test_worker,
        sample_raw_message,
        sample_message,
        mock_instrumentator,
    ):
        test_exception = ValueError("test error")
        test_worker.pull_messages = Mock(return_value=[sample_raw_message])
        test_worker.parse_message = Mock(return_value=sample_message)
        test_worker.work = Mock(side_effect=test_exception)
        test_worker.on_error = Mock()
        
        # Simulate one iteration of the start loop with exception handling
        raw_messages = test_worker.pull_messages()
        messages = [test_worker.parse_message(message=raw_message) for raw_message in raw_messages]
        try:
            test_worker.work(messages=messages)
        except Exception as exception:
            test_worker.logger.exception(
                msg='Failed to process messages',
                extra={'worker_name': test_worker.name},
            )
            mock_instrumentator.increment_counter(
                counter_name='sqs.manager.exceptions',
                attributes={
                    'worker_name': test_worker.name,
                    'error_type': type(exception).__name__,
                },
            )
            test_worker.on_error(exception=exception, messages=messages)
        
        mock_instrumentator.increment_counter.assert_called_once_with(
            counter_name='sqs.manager.exceptions',
            attributes={
                'worker_name': 'test_worker',
                'error_type': 'ValueError',
            },
        )
        test_worker.on_error.assert_called_once_with(
            exception=test_exception,
            messages=[sample_message],
        )

    def test_start_When_pull_messages_raises_exception_Then_continues(
        self,
        test_worker,
        mock_logger,
    ):
        test_worker.pull_messages = Mock(side_effect=Exception("pull error"))
        
        # Simulate one iteration of the start loop with exception handling
        try:
            test_worker.pull_messages()
        except Exception:
            test_worker.logger.exception(msg='Failed to pull messages')
        
        mock_logger.exception.assert_called_with(msg='Failed to pull messages')

    def test_start_When_work_successful_Then_records_latency_metric(
        self,
        test_worker,
        sample_raw_message,
        sample_message,
        mock_instrumentator,
    ):
        test_worker.pull_messages = Mock(return_value=[sample_raw_message])
        test_worker.parse_message = Mock(return_value=sample_message)
        test_worker.work = Mock()
        test_worker.ack_all = Mock()
        
        # Simulate one iteration of the start loop with timing
        with patch('time.time', side_effect=[100.0, 105.0]):
            start_time = time.time()
            raw_messages = test_worker.pull_messages()
            messages = [test_worker.parse_message(message=raw_message) for raw_message in raw_messages]
            test_worker.work(messages=messages)
            
            test_worker.logger.info(
                msg='Worker successfully processed messages',
                extra={'worker_name': test_worker.name},
            )
            mock_instrumentator.record_histogram(
                histogram_name='sqs.manager.work.latency',
                attributes={'worker_name': test_worker.name},
                amount=(time.time() - start_time),
            )
            if test_worker.auto_ack:
                test_worker.ack_all(messages=messages)
        
        mock_instrumentator.record_histogram.assert_called_once_with(
            histogram_name='sqs.manager.work.latency',
            attributes={'worker_name': 'test_worker'},
            amount=5.0,
        )


class TestWorkerStop:
    def test_stop_When_called_Then_sets_active_false(
        self,
        test_worker,
        mock_logger,
    ):
        test_worker.stop()
        
        assert test_worker.active is False
        mock_logger.info.assert_called_once_with(
            msg='Stopping worker',
            extra={'worker_name': 'test_worker'},
        )


class TestWorkerWork:
    def test_work_When_called_Then_raises_not_implemented_error(
        self,
        test_worker,
        sample_message,
    ):
        with pytest.raises(NotImplementedError):
            test_worker.work(messages=[sample_message])


class TestWorkerPullMessages:
    def test_pull_messages_When_called_Then_returns_messages(
        self,
        test_worker,
    ):
        test_worker.get_queue_url = Mock(return_value="https://queue-url")
        # Provide enough empty responses to fill the wait time
        test_worker.sqs_client.receive_message.side_effect = [
            {'Messages': [{'Body': 'test', 'ReceiptHandle': 'handle1'}]},
        ] + [{}] * 10  # Multiple empty responses to avoid StopIteration
        
        # Mock time to control the timeout
        with patch('time.time', side_effect=[0.0, 3.0]):  # Forces timeout after first call
            result = test_worker.pull_messages()
        
        assert len(result) >= 1
        assert result[0]['Body'] == 'test'
        test_worker.sqs_client.receive_message.assert_called()

    def test_pull_messages_When_no_messages_Then_returns_empty_list(
        self,
        test_worker,
    ):
        test_worker.get_queue_url = Mock(return_value="https://queue-url")
        test_worker.sqs_client.receive_message.return_value = {}
        
        with patch('time.time', side_effect=[0.0, 3.0]):
            result = test_worker.pull_messages()
        
        assert result == []

    def test_pull_messages_When_max_messages_reached_Then_stops_polling(
        self,
        test_worker,
    ):
        test_worker.max_messages = 2
        test_worker.get_queue_url = Mock(return_value="https://queue-url")
        test_worker.sqs_client.receive_message.return_value = {
            'Messages': [
                {'Body': 'test1', 'ReceiptHandle': 'handle1'},
                {'Body': 'test2', 'ReceiptHandle': 'handle2'},
            ]
        }
        
        result = test_worker.pull_messages()
        
        assert len(result) == 2
        test_worker.sqs_client.receive_message.assert_called_once()


class TestWorkerParseMessage:
    def test_parse_message_When_valid_message_Then_returns_message_object(
        self,
        test_worker,
        sample_raw_message,
    ):
        result = test_worker.parse_message(sample_raw_message)
        
        assert isinstance(result, models.Message)
        assert result.version == "1.0"
        assert result.msg_type == "test"
        assert result.queue_name == "test_queue"
        assert result.receipt_handle == "receipt_123"
        assert isinstance(result.payload, PayloadForTesting)
        assert result.payload.message == "test message"
        assert result.payload.count == 42
        assert result.retry_attempt == 0

    def test_parse_message_When_message_has_extra_fields_Then_ignores_queue_name_and_receipt_handle(
        self,
        test_worker,
        sample_payload,
    ):
        raw_message = {
            'Body': json.dumps({
                'version': '1.0',
                'msg_type': 'test',
                'queue_name': 'ignored_queue',
                'receipt_handle': 'ignored_handle',
                'payload': sample_payload.model_dump(),
                'retry_attempt': 1,
            }),
            'ReceiptHandle': 'receipt_456',
        }
        
        result = test_worker.parse_message(raw_message)
        
        assert result.queue_name == "test_queue"
        assert result.receipt_handle == "receipt_456"
        assert result.retry_attempt == 1


class TestWorkerRetryLater:
    def test_retry_later_When_messages_within_retry_limit_Then_requeues_messages(
        self,
        test_worker,
        sample_message,
        mock_logger,
    ):
        sample_message.retry_attempt = 2
        test_worker.publish_messages = Mock()
        test_worker.ack_all = Mock()
        
        test_worker.retry_later(messages=[sample_message])
        
        assert sample_message.retry_attempt == 3
        test_worker.publish_messages.assert_called_once_with(
            queue_name="test_queue",
            messages=[sample_message],
            delay_seconds=datetime.timedelta(minutes=10),
        )
        test_worker.ack_all.assert_called_once_with(messages=[sample_message])
        mock_logger.info.assert_called_once_with(
            msg='Enqueuing retry later messages',
            extra={'messages_count': 1},
        )

    def test_retry_later_When_messages_exceed_retry_limit_Then_skips_messages(
        self,
        test_worker,
        sample_message,
    ):
        sample_message.retry_attempt = 5
        test_worker.max_retry_attempts = 3
        test_worker.publish_messages = Mock()
        test_worker.ack_all = Mock()
        
        test_worker.retry_later(messages=[sample_message])
        
        test_worker.publish_messages.assert_called_once_with(
            queue_name="test_queue",
            messages=[],
            delay_seconds=datetime.timedelta(minutes=10),
        )
        test_worker.ack_all.assert_called_once_with(messages=[])


class TestWorkerPublishMessage:
    def test_publish_message_When_called_Then_sends_message_to_queue(
        self,
        test_worker,
        sample_message,
    ):
        test_worker.get_queue_url = Mock(return_value="https://queue-url")
        
        test_worker.publish_message(
            queue_name="test_queue",
            message=sample_message,
        )
        
        test_worker.sqs_client.send_message.assert_called_once_with(
            QueueUrl="https://queue-url",
            MessageBody=sample_message.model_dump_json(),
        )


class TestWorkerPublishMessages:
    def test_publish_messages_When_empty_list_Then_returns_early(
        self,
        test_worker,
    ):
        test_worker.get_queue_url = Mock()
        
        test_worker.publish_messages(
            queue_name="test_queue",
            messages=[],
        )
        
        test_worker.get_queue_url.assert_not_called()
        test_worker.sqs_client.send_message_batch.assert_not_called()

    def test_publish_messages_When_single_batch_Then_sends_batch(
        self,
        test_worker,
        sample_message,
    ):
        test_worker.get_queue_url = Mock(return_value="https://queue-url")
        
        test_worker.publish_messages(
            queue_name="test_queue",
            messages=[sample_message],
        )
        
        expected_entries = [{
            'Id': 'message-0',
            'MessageBody': sample_message.model_dump_json(),
        }]
        test_worker.sqs_client.send_message_batch.assert_called_once_with(
            QueueUrl="https://queue-url",
            Entries=expected_entries,
        )

    def test_publish_messages_When_with_delay_Then_adds_delay_seconds(
        self,
        test_worker,
        sample_message,
    ):
        test_worker.get_queue_url = Mock(return_value="https://queue-url")
        delay = datetime.timedelta(minutes=5)
        
        test_worker.publish_messages(
            queue_name="test_queue",
            messages=[sample_message],
            delay_seconds=delay,
        )
        
        expected_entries = [{
            'Id': 'message-0',
            'MessageBody': sample_message.model_dump_json(),
            'DelaySeconds': 300,
        }]
        test_worker.sqs_client.send_message_batch.assert_called_once_with(
            QueueUrl="https://queue-url",
            Entries=expected_entries,
        )

    def test_publish_messages_When_multiple_batches_Then_sends_multiple_batches(
        self,
        test_worker,
        sample_payload,
    ):
        messages = [
            models.Message(
                version="1.0",
                msg_type="test",
                payload=sample_payload,
            )
            for _ in range(25)
        ]
        test_worker.get_queue_url = Mock(return_value="https://queue-url")
        
        test_worker.publish_messages(
            queue_name="test_queue",
            messages=messages,
        )
        
        assert test_worker.sqs_client.send_message_batch.call_count == 3


class TestWorkerAck:
    def test_ack_When_called_Then_calls_ack_all(
        self,
        test_worker,
        sample_message,
    ):
        test_worker.ack_all = Mock()
        
        test_worker.ack(sample_message)
        
        test_worker.ack_all.assert_called_once_with(messages=[sample_message])


class TestWorkerAckAll:
    def test_ack_all_When_messages_have_receipt_handles_Then_deletes_messages(
        self,
        test_worker,
        sample_message,
    ):
        test_worker.get_queue_url = Mock(return_value="https://queue-url")
        
        test_worker.ack_all(messages=[sample_message])
        
        expected_entries = [{
            'Id': 'message-0',
            'ReceiptHandle': 'receipt_123',
        }]
        test_worker.sqs_client.delete_message_batch.assert_called_once_with(
            QueueUrl="https://queue-url",
            Entries=expected_entries,
        )

    def test_ack_all_When_messages_without_receipt_handles_Then_skips_messages(
        self,
        test_worker,
        sample_payload,
    ):
        message_without_handle = models.Message(
            version="1.0",
            msg_type="test",
            payload=sample_payload,
            receipt_handle=None,
        )
        test_worker.get_queue_url = Mock(return_value="https://queue-url")
        
        test_worker.ack_all(messages=[message_without_handle])
        
        test_worker.sqs_client.delete_message_batch.assert_called_once_with(
            QueueUrl="https://queue-url",
            Entries=[],
        )


class TestWorkerExtendMessageVisibility:
    def test_extend_message_visibility_When_valid_message_Then_changes_visibility(
        self,
        test_worker,
        sample_message,
    ):
        test_worker.get_queue_url = Mock(return_value="https://queue-url")
        
        test_worker.extend_message_visibility(
            message=sample_message,
            visibility_timeout=60,
        )
        
        test_worker.sqs_client.change_message_visibility.assert_called_once_with(
            QueueUrl="https://queue-url",
            ReceiptHandle="receipt_123",
            VisibilityTimeout=60,
        )

    def test_extend_message_visibility_When_no_receipt_handle_Then_raises_value_error(
        self,
        test_worker,
        sample_payload,
    ):
        message_without_handle = models.Message(
            version="1.0",
            msg_type="test",
            payload=sample_payload,
            receipt_handle=None,
        )
        
        with pytest.raises(ValueError, match="Unable to extend visibility without receipt_handle"):
            test_worker.extend_message_visibility(
                message=message_without_handle,
                visibility_timeout=60,
            )


class TestWorkerGetQueueUrl:
    def test_get_queue_url_When_called_Then_returns_queue_url(
        self,
        test_worker,
    ):
        test_worker.sqs_client.get_queue_url.return_value = {
            'QueueUrl': 'https://queue-url'
        }
        
        result = test_worker.get_queue_url(queue_name="test_queue")
        
        assert result == "https://queue-url"
        test_worker.sqs_client.get_queue_url.assert_called_once_with(
            QueueName="test_queue"
        )


class TestWorkerOnError:
    def test_on_error_When_called_Then_does_nothing(
        self,
        test_worker,
        sample_message,
    ):
        exception = ValueError("test error")
        
        result = test_worker.on_error(
            exception=exception,
            messages=[sample_message],
        )
        
        assert result is None


class TestSingle:
    def test_single_When_called_Then_calls_multiple_with_n_equals_1(
        self,
    ):
        worker_factory = Mock()
        
        with patch.object(worker, 'multiple') as mock_multiple:
            worker.single(worker_factory)
        
        mock_multiple.assert_called_once_with(
            worker_factory=worker_factory,
            n=1,
        )


class TestMultiple:
    def test_multiple_When_called_Then_creates_workers_and_starts_them(
        self,
        test_worker,
    ):
        worker_factory = Mock(return_value=test_worker)
        
        with patch('concurrent.futures.ThreadPoolExecutor') as mock_executor_class:
            with patch('signal.signal') as mock_signal:
                mock_executor = Mock()
                mock_executor_class.return_value = mock_executor
                
                worker.multiple(
                    worker_factory=worker_factory,
                    n=2,
                )
                
                assert worker_factory.call_count == 2
                mock_executor.submit.assert_called()
                mock_signal.assert_called_once_with(
                    signal.SIGTERM,
                    ANY,
                )

    def test_multiple_When_sigterm_received_Then_stops_workers(
        self,
    ):
        mock_worker1 = Mock()
        mock_worker2 = Mock()
        worker_factory = Mock(side_effect=[mock_worker1, mock_worker2])
        captured_handler = None
        
        def capture_signal_handler(
            sig,
            handler,
        ):
            nonlocal captured_handler
            captured_handler = handler
        
        with patch('concurrent.futures.ThreadPoolExecutor'):
            with patch('signal.signal', side_effect=capture_signal_handler):
                worker.multiple(
                    worker_factory=worker_factory,
                    n=2,
                )
                
                captured_handler(signal.SIGTERM, None)
                
                mock_worker1.stop.assert_called_once()
                mock_worker2.stop.assert_called_once()



import mypy_boto3_sqs.client
import sqs_worker


def sqs_connected(
    sqs_client: mypy_boto3_sqs.client.SQSClient,
) -> bool:
    try:
        sqs_client.list_queues()
        return True
    except Exception:
        return False


def ensure_queue_exists(
    sqs_client: mypy_boto3_sqs.client.SQSClient,
    queue_name: str,
) -> None:
    try:
        sqs_client.get_queue_url(QueueName=queue_name)
    except sqs_client.exceptions.QueueDoesNotExist:
        sqs_client.create_queue(QueueName=queue_name)


def send_hello_message(
    sqs_client: mypy_boto3_sqs.client.SQSClient,
    queue_name: str,
    message: sqs_worker.models.Message,
) -> None:
    queue_url = sqs_client.get_queue_url(QueueName=queue_name)['QueueUrl']
    sqs_client.send_message(
        QueueUrl=queue_url,
        MessageBody=message.model_dump_json(),
    )


"""SQS Worker - Manage workers triggered by AWS SQS messages."""

__version__ = "0.1.0"

from . import exceptions
from . import models
from . import worker

__all__ = ["exceptions", "models", "worker"]